/*//////////////////////////////////////////////////////////////////////////////////////////////////
///  HOCRdemo.cpp    Higher-Order Clique Reduction Denoising Demo
///  Version 1.01
////////////////////////////////////////////////////////////////////////////////////////////////////

Copyright 2009 Hiroshi Ishikawa.
This software can be used for research purposes only.
This software or its derivatives must not be publicly distributed
without a prior consent from the author (Hiroshi Ishikawa).

THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

For the latest version, check: http://www.nsc.nagoya-cu.ac.jp/~hi/

////////////////////////////////////////////////////////////////////////////////////////////////////

Loads the noise-added image and denoise it with a third-order FoE prior.
The experiment in this demo is described in the following paper:

Hiroshi Ishikawa. "Higher-Order Clique Reduction in Binary Graph Cut."
In CVPR2009: IEEE Computer Society Conference on Computer Vision and Pattern Recognition,
Miami Beach, Florida. June 20-25, 2009.

FoE is described in the following paper:

Stefan Roth and Michael J. Black. "Fields of Experts: A Framework for Learning Image Priors."
In CVPR2005: IEEE Computer Society Conference on Computer Vision and Pattern Recognition,
San Diego, California, June 20-25, 2005, p.p. II:860-867.

This software requires the QPBO software by Vladimir Kolmogorov available
at http://www.cs.ucl.ac.uk/staff/V.Kolmogorov/software.html

The code was tested on Microsoft Visual C++ 2008 Express Edition SP1
and gcc version 4.1.2.
Any report on bugs and results of trying on other platforms is appreciated.

//////////////////////////////////////////////////////////////////////////////////////////////////*/

#include "HOCR/HOCR.h"
#include "QPBO/QPBO.h"
#include "Image.h"

using namespace std;

// Image filenames
const char* NOISE_ADDED = "test001_020.pgm"; // The given noisy image.
const char* ORIGINAL = "test001.pgm"; // For PSNR calculation
const char* RESULT = "test001_020_result.pgm";  // To save the denoised image.

// Parameters
typedef int REAL;	// Type for the energy value
const REAL maxIt = 300;	// Maximum number of iteration.
const REAL sigma = 20;	// Sigma of the noise added
const int mult = 10000;	// Multiplier for representing real values by integers
const int Ecycle = 20;	// Number of iterations to check energy decrease.
const double stopThreashold = 100.0; // Stop if the energy has changed less than this after Ecycle iterations. 
const int Bcycle = 30;	// Number of iterations to renew the blurred image.

// direction vectors
const int vx[4] = {0, 1, 0, 1};
const int vy[4] = {0, 0, 1, 1};

// FoE experts provided by Stefan Roth 
double alpha[3] = {0.586612685392731, 1.157638405566669, 0.846059486257292};
double expert[3][4] = {
	{-0.0582774013402734, 0.0339010363051084, -0.0501593018104054, 0.0745568557931712},
	{0.0492112815304123, -0.0307820846538285, -0.123247230948424, 0.104812330861557},
	{0.0562633568728865, 0.0152832583489560, -0.0576215592718086, -0.0139673758425540}
};


template<typename T> T square(const T& t) {return t * t;}


double getFoELocal(unsigned int h[4])
{
	double e = 0;
	for (int j = 0; j < 3; j++)
	{
		double s = 0;
		for (int k = 0; k < 4; k++)
			s += expert[j][k] * h[k];
		e += alpha[j] * log(1 + 0.5 * square(s));
	}
	return e;
}


double unaryEnergy(int x, int data, int sigma)
{
	return (double)square(x - data) / (square(sigma) * 2);
}


template<typename F>
void addEnergy(F& f, const image& cur, const image& im, const image& pro, int W, int H, int N, REAL mult, REAL sigma)
{
	for (int y = 0; y + 1 < H; y++)	// FoE prior
		for (int x = 0; x + 1 < W; x++) 
		{
			int nds[4];
			for (int j = 0; j < 4; j++)
				nds[j] = x + vx[j] + (y + vy[j]) * W;
			REAL E[16];
			for (int i = 0; i < 16; i++) // i = 0000 (binary) means all current, i = 1000 means only node 0 is proposed, 
			{                                      // i = 1001 means node 0 and 3 are proposed, etc.
				unsigned int h[4];
				int b = 8;
				for (int j = 0; j < 4; j++)
				{
					h[j] = (i & b) ? pro.buf[nds[j]] : cur.buf[nds[j]];
					b >>= 1;
				}
				E[i] = (REAL)(getFoELocal(h) * mult);
			}
			f.AddHigherTerm(4, nds, E);
		}
	for (int j = 0; j < N; j++) // Data term
	{
		double e0 = unaryEnergy(cur.buf[j], im.buf[j], sigma);
		double e1 = unaryEnergy(pro.buf[j], im.buf[j], sigma);
		f.AddUnaryTerm(j, (REAL)(e0 * mult), (REAL)(e1 * mult));
	}
}

// buildQPBF1 and buildQPBF2: Two alternatives to use HOCR software. See HOCR.h

void buildQPBF1(QPBO<REAL>& qpbo, const image& cur, const image& im, const image& pro, int W, int H, int N, REAL mult, REAL sigma)
{
	HOCR<REAL,4,QPBO<REAL> > hocr(qpbo);
	hocr.AddNode(N);
	addEnergy(hocr, cur, im, pro, W, H, N, mult, sigma); // Prior
}


void buildQPBF2(QPBO<REAL>& qpbo, const image& cur, const image& im, const image& pro, int W, int H, int N, REAL mult, REAL sigma)
{
	PBF<REAL,4> pbf;
	addEnergy(pbf, cur, im, pro, W, H, N, mult, sigma); // Prior
	PBF<REAL,2> qpbf;
	pbf.toQuadratic(qpbf); // Reduce to Quadatic pseudo-Boolean function
	pbf.clear(); // free memory
	convert(qpbo, qpbf); // copy to QPBO object by V. Kolmogorov
}


double getEnergy(image& f, const image& data, int W, int H, REAL sigma)
{
	double E = 0;
	unsigned int h[4];
	for (int y = 0; y + 1 < H; y++)
		for (int x = 0; x + 1 < W; x++)
		{
			for (int j = 0; j < 4; j++)
				h[j] = f(x + vx[j], y + vy[j]);
			E += getFoELocal(h);
		}
	for (int j = 0; j < H * W; j++)
		E += unaryEnergy(f.buf[j], data.buf[j], sigma);
	return E;
}


double getPSNR(const image& d, const image& o, int N)
{
	int s = 0;
	for (int i = 0; i < N; i++)
		s += square(d.buf[i] - o.buf[i]);
	return 20 * log10(255.0 / sqrt((double)s / N));
}


double rand01()
{
	return ((double)rand()-0.00000000001) / RAND_MAX;
}



int main()
{
	double Erec[Ecycle];
	printf("Original image: %s\nNoise-added image: %s\n", ORIGINAL, NOISE_ADDED);
	image im, org, blr;
	im.readPGM(NOISE_ADDED);
	org.readPGM(ORIGINAL);
	if (im.empty() || org.empty())
	{
		printf("Error. Cannot load the image.\n");
		return 0;
	}
	int W = im.W, H = im.H, N = W * H;
	printf("Initial Energy = %.1f   Initial PSNR = %.4f\n", (double)getEnergy(im, im, W, H, sigma), getPSNR(im, org, N));
	image cur(im);
	image pro(W, H);
	for (int i = 0; i < maxIt; i++)
	{
		if ((i % Bcycle) == 0)
			cur.gaussianblur(blr, .5625);
		if ((i % 2) == 0)
			for (int j = 0; j < N; j++)
				pro.buf[j] = min((REAL)255, max((REAL)0, (REAL)blr.buf[j] + (REAL)((rand01()-0.5) * sigma * 3)));
		else
			for (int j = 0; j < N; j++)
				pro.buf[j] = (int)(rand01() * 256);
		
		QPBO<REAL> qpbo(N * 10, N * 20);
		buildQPBF1(qpbo, cur, im, pro, W, H, N, mult, sigma); // Two alternatives to use HOCR software.
//		buildQPBF2(qpbo, cur, im, pro, W, H, N, mult, sigma); // See HOCR.h
		qpbo.MergeParallelEdges();
		qpbo.Solve();
		qpbo.ComputeWeakPersistencies();
		int labeled = 0;
		for (int j = 0; j < N; j++)
		{
			int res = qpbo.GetLabel(j);
			if (res == 1)
				cur.buf[j] = pro.buf[j];
			if (res >= 0)
				labeled++;
		}
		double psnr = getPSNR(cur, org, N);
		double te = getEnergy(cur, im, W, H, sigma);
		printf("%d (%3.3f%% labeled):  E = %.1f  PSNR = %.4f\n", i, labeled * 100.0 / N, te, psnr);
		if (i >= Ecycle && Erec[i % Ecycle] - te < stopThreashold)
			break;
		Erec[i % Ecycle] = te;
	}
	cur.writePGM(RESULT);
	printf("Denoised image written to %s\n", RESULT);
	return 0;
}



